import pickle
from pydantic import BaseModel
from bondai.tools import Tool
from googleapiclient.discovery import build


MAX_RESULT_COUNT = 20
DEFAULT_RESULT_COUNT = 5
TOOL_NAME = 'list_emails'
TOOL_DESCRIPTION = f"This tool allows you to retrieve a paginated list of emails from the user's inbox that includes the 'Id', 'From' and 'Subject' fields for each email. You can specify the number of items to return by setting the 'count' parameter. The maximum count is {MAX_RESULT_COUNT} and the default is {DEFAULT_RESULT_COUNT}. To paginate through the full list of all emails just increment the 'page' parameter. By default 'page' is set to 1. You can also optionally specificy a valid gmail query in the 'query' parameter."

class Parameters(BaseModel):
    page: int
    count: int
    thought: str

def get_email_attr(message, attr):
    return next((h['value'] for h in message['payload']['headers'] if h['name'] == attr), None)

class ListEmailsTool(Tool):
    def __init__(self, credentials=None, credentials_filename='gmail-token.pickle'):
        super(ListEmailsTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        if credentials:
            self.service = build('gmail', 'v1', credentials=credentials)
        elif credentials_filename:
            with open(credentials_filename, 'rb') as token:
                credentials = pickle.load(token)
        
            self.service = build('gmail', 'v1', credentials=credentials)
        else:
            raise Exception('No credentials provided.')
    
    def run(self, arguments):
        page = int(arguments.get('page', '1'))
        count = int(arguments.get('count', DEFAULT_RESULT_COUNT))
        query = arguments.get('query', '')

        if count > MAX_RESULT_COUNT:
            count = MAX_RESULT_COUNT

        result_size_estimate = None
        page_token = None
        results = {}
        for i in range(page):
            if page == 1:
                results = self.service.users().messages().list(userId='me', labelIds=['INBOX'], maxResults=MAX_RESULT_COUNT, q=query).execute()
            elif page_token:
                results = self.service.users().messages().list(userId='me', labelIds=['INBOX'], maxResults=MAX_RESULT_COUNT, q=query, pageToken=page_token).execute()
            else:
                break
            
            page_token = results.get('nextPageToken', None)
            if not result_size_estimate:
                result_size_estimate = results['resultSizeEstimate']

        messages = results.get('messages', [])

        if len(results) > 0:
            output = f"The total number of messages: {result_size_estimate}\n\n"
            for m in messages:
                msg_id = m['id']
                message = self.service.users().messages().get(userId='me', id=msg_id).execute()
                msg_from = get_email_attr(message, 'From')
                msg_received = get_email_attr(message, 'Date')
                msg_subj = get_email_attr(message, 'Subject')
                output += f"Id: {msg_id}\nDate: {msg_received}\nFrom: {msg_from}\nSubject: {msg_subj}\n\n"
            
            print(output)
            return output
        else:
            return '0 messages were found.'




        


    




MODEL_TYPE_LLM = 'MODEL_TYPE_LLM'
MODEL_TYPE_EMBEDDING = 'MODEL_TYPE_EMBEDDING'

MODEL_GPT4 = 'gpt-4'
MODEL_GPT4_0613 = 'gpt-4-0613'
MODEL_GPT4_32K = 'gpt-4-32k'
MODEL_GPT35_TURBO = 'gpt-3.5-turbo'
MODEL_GPT35_TURBO_16K = 'gpt-3.5-turbo-16k'
MODEL_GPT35_TURBO_0613 = 'gpt-3.5-turbo-0613'
MODEL_GPT35_TURBO_16K_0613 = 'gpt-3.5-turbo-16k-0613'
MODEL_TEXT_EMBEDDING_ADA_002 = 'text-embedding-ada-002'

MODEL_NAMES = [
    MODEL_GPT4,
    MODEL_GPT4_0613,
    MODEL_GPT4_32K,
    MODEL_GPT35_TURBO,
    MODEL_GPT35_TURBO_16K,
    MODEL_GPT35_TURBO_0613,
    MODEL_GPT35_TURBO_16K_0613,
    MODEL_TEXT_EMBEDDING_ADA_002,
]

MODELS = {
    MODEL_GPT4: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 8192,
        'input_price_per_token': 0.00003,
        'output_price_per_token': 0.00006,
    },
    MODEL_GPT4_0613: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 8192,
        'input_price_per_token': 0.00003,
        'output_price_per_token': 0.00006,
    },
    MODEL_GPT4_32K: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 32768,
        'input_price_per_token': 0.00006,
        'output_price_per_token': 0.00012,
    },
    MODEL_GPT35_TURBO: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 4096 ,
        'input_price_per_token': 0.0000015,
        'output_price_per_token': 0.000002,
    },
    MODEL_GPT35_TURBO_16K: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 16384 ,
        'input_price_per_token': 0.000003,
        'output_price_per_token': 0.000004,
    },
    MODEL_GPT35_TURBO_0613: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 4096 ,
        'input_price_per_token': 0.0000015,
        'output_price_per_token': 0.000002,
    },
    MODEL_GPT35_TURBO_16K_0613: {
        'model_type': MODEL_TYPE_LLM,
        'max_tokens': 16384 ,
        'input_price_per_token': 0.000003,
        'output_price_per_token': 0.000004,
    },
    MODEL_TEXT_EMBEDDING_ADA_002: {
        'model_type': MODEL_TYPE_EMBEDDING,
        'max_tokens': 8191 ,
        'price_per_token': 0.0000001
    },
}


import pickle
import base64
from googleapiclient.discovery import build
from typing import List
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.util import get_html_text
from bondai.models.openai import OpenAILLM, MODEL_GPT35_TURBO_16K

TOOL_NAME = 'get_email_content'
QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about emails. Use the provided information to answer the user's QUESTION at the very end."
TOOL_DESCRIPTION = "This tool allows to ask a question about the text content of a list of emails including summarization. Simply provide a comma seperated list of email ids in the 'email_ids' parameter and specify your question using the 'question' parameter."

def get_email_attr(message, attr):
    return next((h['value'] for h in message['payload']['headers'] if h['name'] == attr), None)

def build_prompt(question, context):
    return f"""{context}


IMPORTANT: Answer the following question for the user.
QUESTION: {question}
"""

def parse_body(message):
    payload = message['payload']

    if 'parts' in payload:
        htmlPart = next((p for p in payload['parts'] if p['mimeType'] == 'text/html'), None)
        textPart = next((p for p in payload['parts'] if p['mimeType'] == 'text/plain'), None)
    elif 'mimeType' in payload:
        if payload['mimeType'] == 'text/html':
            htmlPart = payload
        elif payload['mimeType'] == 'text/plain':
            textPart = payload

    if htmlPart:
        data = htmlPart['body']['data'].replace("-","+").replace("_","/")
        decoded_data = base64.b64decode(data)
        return get_html_text(decoded_data)
    elif textPart:
        data = textPart['body']['data'].replace("-","+").replace("_","/")
        return base64.b64decode(data)

class Parameters(BaseModel):
    email_ids: List[str]
    question: str
    thought: str

class QueryEmailsTool(Tool):
    def __init__(self, credentials=None, credentials_filename='gmail-token.pickle', llm=OpenAILLM(MODEL_GPT35_TURBO_16K)):
        super(QueryEmailsTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.llm = llm
        if credentials:
            self.service = build('gmail', 'v1', credentials=credentials)
        elif credentials_filename:
            with open(credentials_filename, 'rb') as token:
                credentials = pickle.load(token)
        
            self.service = build('gmail', 'v1', credentials=credentials)
        else:
            raise Exception('No credentials provided.')
    
    def run(self, arguments):
        question = arguments.get('question')
        email_ids = arguments.get('email_ids')

        if question is None:
            raise Exception('question is required')
        if email_ids is None:
            raise Exception('email_ids is required')

        if isinstance(email_ids, str):
            email_ids = map(lambda id: id.strip(), email_ids.split(','))

        text = ''
        for msg_id in email_ids:
            message = self.service.users().messages().get(userId='me', id=msg_id).execute()
            msg_from = get_email_attr(message, 'From')
            msg_subj = get_email_attr(message, 'Subject')
            msg_body = parse_body(message)
            text += f"Id: {msg_id}\nFrom: {msg_from}\nSubject: {msg_subj}\nBody: {msg_body}\n\n"
        
        prompt = build_prompt(question, text)
        response = self.llm.get_completion(prompt, QUERY_SYSTEM_PROMPT)[0]

        return response




def format_order_response(response):
    return f"""Order ID: {response.id}
Status: {response.status}
Filled At: {response.filled_at}
Symbol: {response.symbol}
Asset Class: {response.asset_class}
Notional: {response.notional}
Quantity: {response.qty}
Filled Quantity: {response.filled_qty}
Filled Average Price: {response.filled_avg_price}
Filled Price: {response.filled_avg_price}
Order Class: {response.order_class}
Order Type: {response.order_type}
Side: {response.side}
Time In Force: {response.time_in_force}
Limit Price: {response.limit_price}
Stop Price: {response.stop_price}"""

def format_account_response(response):
    return f"""Cash: {response.cash}
Currency: {response.currency}
Buying Power: {response.buying_power}
Regt Buying Power: {response.regt_buying_power}
Daytrading Buying Power: {response.daytrading_buying_power}  
Non Marginable Buying Power: {response.non_marginable_buying_power}
Accrued Fees: {response.accrued_fees}
Portfolio Value: {response.portfolio_value}
Shorting Enabled: {response.shorting_enabled}
Crypto Status: {response.crypto_status}
Long Market Value: {response.long_market_value}
Short Market Value: {response.short_market_value}
Initial Margin: {response.initial_margin}
Maintenance Margin: {response.maintenance_margin}
Last Maintenance Margin: {response.last_maintenance_margin}"""

def format_positions_response(response):
    result = ''
    for position in response:
        result += format_position(position) + '\n\n'
    return result

def format_position(position):
    return f"""Symbol: {position.symbol}
Quantity: {position.qty}
Entry Share Price: {position.avg_entry_price}
Current Share Price: {position.current_price}
Unrealized Profit/Loss: {position.unrealized_pl}
Market Value: {position.market_value}
Cost Basis: {position.cost_basis}"""

def format_orders_response(response):
    result = ''
    for order in response:
        result += format_order(order) + '\n\n'
    return result

def format_order(order):
    return f"""Order ID: {order.id}
Type: {order.type}
Symbol: {order.symbol}
Status: {order.status}
Side: {order.side}
Limit Price: {order.limit_price}
Quantity: {order.qty}"""



import os
from bondai.tools import Tool
from .response_formatter import format_account_response
from alpaca.trading.client import TradingClient

TOOL_NAME = 'get_account_info'
TOOL_DESCRIPTION = "This tool will provide information about your investment account including your cash balance."

ALPACA_API_KEY = os.environ.get('ALPACA_API_KEY')
ALPACA_SECRET_KEY = os.environ.get('ALPACA_SECRET_KEY')

class GetAccountTool(Tool):
    def __init__(self, alpaca_api_key=ALPACA_API_KEY, alpaca_secret_key=ALPACA_SECRET_KEY):
        super(GetAccountTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION)
        self.trading_client = TradingClient(alpaca_api_key, alpaca_secret_key, paper=True)
    
    def run(self, arguments):
        response = self.trading_client.get_account()
        return format_account_response(response)




from abc import ABC, abstractmethod

class LLM(ABC):

    @abstractmethod
    def get_completion(prompt, system_prompt='', previous_messages=[], functions=[]):
        pass

    @abstractmethod
    def count_tokens(prompt):
        pass

    @abstractmethod
    def get_max_tokens():
        pass


import json
import traceback
from termcolor import colored, cprint
from bondai.tools.tool import Tool, InputParameters
from bondai.tools.response_query import ResponseQueryTool
from bondai.prompt import DefaultPromptBuilder
from bondai.models.openai import (
    OpenAILLM, 
    MODEL_GPT4_0613, 
    MODEL_GPT35_TURBO_0613, 
    get_total_cost
)

TOOL_MAX_TOKEN_RESPONSE = 2000
DEFAULT_FINAL_ANSWER_TOOL = Tool('final_answer', "Use the final_answer tool when you have all the information you need to provide the final answer.", InputParameters)

def format_print_string(s, length=100):
    # Remove newlines
    s = s.replace('\n', ' ').replace('\r', '')
    
    if len(s) <= length:
        return s
    return s[:length - 3] + "..."

def load_alternative_function(llm, function, options):
    args = {}
    thought = ''
    if 'arguments' in function:
        try:
            args = json.loads(function['arguments'])
            if 'thought' in args:
                thought = "Thought: " + args['thought'] + "\n"
        except:
            pass
    
    function_descriptions = '\n'.join([f"{tool['name']: {tool['description']}}" for tool in options])
    prompt = f"""An AI agent is attempting to use a function that does not exist. Please determine if one of the options provided is a good alternative.
Incorrect Function: {function['name']}
{thought}

VERY IMPORTANT:
- If none of the options is a match simpley return "NO MATCH" as the output.

Options:
id: description
{function_descriptions}
"""
    _, new_function = llm.get_completion(prompt, options)

    result = next((f for f in options if f.name == new_function['name']), None)

    return result


class AgentStep:

    def __init__(self, prompt, message, function=None):
        self.prompt = prompt
        self.message = message
        self.function = function
        self.output = None
        self.error = False
        self.final_answer = False
        self.monitor_feedback = None

class BudgetExceededException(Exception):
    pass

class Agent:

    def __init__(self, prompt_builder=None, tools=[], llm=OpenAILLM(MODEL_GPT4_0613), fallback_llm=OpenAILLM(MODEL_GPT35_TURBO_0613), final_answer_tool=DEFAULT_FINAL_ANSWER_TOOL, budget=None, quiet=False, additional_tools=[]):
        if not prompt_builder:
            self.prompt_builder = DefaultPromptBuilder(llm)
        else:
            self.prompt_builder = prompt_builder
        
        self.previous_steps = []
        self.response_query_tool = ResponseQueryTool()
        self.tools = tools
        self.llm = llm
        self.fallback_llm = fallback_llm
        self.final_answer_tool = final_answer_tool
        self.budget = budget
        self.quiet = quiet
        self.additional_tools = additional_tools
    
    def run_once(self, task=''):
        tools = self.tools
        if len(self.response_query_tool.responses) > 0:
            tools.append(self.response_query_tool)

        prompt = self.prompt_builder.build_prompt(task, tools, self.previous_steps)
        functions = list(map(lambda t: t.get_tool_function(), tools))

        message, function = self.llm.get_completion(prompt, '', [], functions)

        if not function:
            message, function = self.fallback_llm.get_completion(message, system_prompt='Select the correct function and parameters to use based on the prompt', previous_messages=[], functions=functions)

        if not function and not self.quiet:
            cprint(f"No function was selected.", 'red')
            cprint(message, 'red')
            
        if function:
            step = AgentStep(prompt, message, function)
            tools = [t for t in self.tools if t.name == function['name']]
            if len(tools) > 0:
                tool = tools[0]
                
                try:
                    args = {}
                    if 'arguments' in function:
                        args = json.loads(function['arguments'])
                    
                    if not self.quiet:
                        cprint(f"\n\nUsing the {tool.name} tool", 'white', attrs=["bold"])
                        if len(args) > 0:
                            if 'thought' in args:
                                print(colored("Thought:", 'white', attrs=["bold"]), colored(args['thought'], 'white'))

                            cprint("Arguments", 'white', attrs=["bold"])
                            for key, value in args.items():
                                if key != 'thought':
                                    cprint(f"{key}: {format_print_string(str(value))}", 'white')


                    try:
                        step.output = tool.run(args)
                        step.final_answer = tool.name == self.final_answer_tool.name
                        
                        if step.output:
                            if self.llm.count_tokens(step.output) > TOOL_MAX_TOKEN_RESPONSE:
                                response_id = self.response_query_tool.add_response(step.output)
                                step.output = f"The result from this tool was greater than {TOOL_MAX_TOKEN_RESPONSE} tokens and could not be displayed. However, you can use the response_query tool to ask questions about the content of this response. Just use response_id = {response_id}."
                        else:
                            step.output = "This command ran successfully with no output."
                        
                        if not self.quiet:
                            print(colored("Output:", 'white', attrs=["bold"]), colored(format_print_string(step.output), 'white'))
                    except Exception as e:
                        # traceback.print_exc()
                        if not self.quiet:
                            cprint(f"An Error occured: {e}", 'red', attrs=["bold"])
                        step.error = True
                        step.output = "An Error occured: " + str(e)
                except Exception as e:
                    if not self.quiet:
                        cprint(f"An Error occured: {e}", 'red', attrs=["bold"])
                    step.error = True
                    step.output = 'An Error occured: The provided arguments were not valid JSON. Valid JSON must ALWAYS BE PROVIDED in the response.'
            else:
                step.error = True
                tool_added = False

                if not self.quiet:
                    cprint(f"No tools found for: {function['name']}", 'red', attrs=["bold"])
                
                new_function = load_alternative_function(self.fallback_llm, function, self.additional_tools)
                if new_function:
                    if new_function.dangerous:
                        cprint(f"The BondAI would like to load a new function: {new_function['name']}. This function is considered dangerous. Would you like to allow BondAI to load this function? (yes/no)\n\n", 'red', attrs=["bold"])
                        answer = input()
                        if answer == 'yes':
                            self.tools.append(new_function)
                            tool_added = True
                    else:
                        self.tools.append(new_function)
                        tool_added = True

                if tool_added:
                    cprint(f"Alternate tool found: {new_function['name']}", 'yellow', attrs=["bold"])
                    step.output = f"An Error occured: The function '{function['name']}' does not exist but an alterantive tool was found and will be loaded for future use: {new_function['name']}."
                else:
                    step.output = f"An Error occured: The function '{function['name']}' does not exist."
                
                
        else:
            step = AgentStep(prompt, message)
            step.output = "No function was provided."
        
        return step
    
    def reset_memory(self):
        self.previous_steps = []

    def run(self, task=''):
        self.tools = self.tools + [self.final_answer_tool]

        while True:
            step = self.run_once(task)

            if step.function or len(self.previous_steps) == 0 or self.previous_steps[-1].function:
                self.previous_steps.append(step)
            
            total_cost = get_total_cost()
            if not self.quiet:
                print(colored("Total Cost:", 'green', attrs=["bold"]), colored(str(round(total_cost, 2)), 'white'))

            if step.final_answer:
                return step

            if self.budget and total_cost >= self.budget:
                raise BudgetExceededException("The budget for this task has been reached without successfully finishing.")





from pydantic import BaseModel
import pkg_resources
from bondai.tools import LangChainTool

class ShellToolParameters(BaseModel):
    commands: str
    thought: str

class PythonREPLParameters(BaseModel):
    query: str

def is_langchain_installed():
    try:
        pkg_resources.get_distribution("langchain")
        return True
    except pkg_resources.DistributionNotFound:
        return False

def build_langchain_tools():
    """Build Langchain tools."""
    from langchain.tools import ShellTool
    from langchain.tools.python.tool import PythonREPLTool

    shell_tool = ShellTool()
    shell_tool.description = "This is a very powerful tool that allows you to run any shell command on this Ubuntu machine. Note that this tool only accepts a single string argument at a time and does not accept a list of commands." + f" args {shell_tool.args}".replace(
        "{", "{{"
    ).replace("}", "}}")

    return [
        LangChainTool(shell_tool, ShellToolParameters, dangerous=True), 
        LangChainTool(PythonREPLTool(), PythonREPLParameters, dangerous=True)
    ]

    






from pydantic import BaseModel

class InputParameters(BaseModel):
    input: str
    thought: str

class EmptyParameters(BaseModel):
    thought: str

class Tool():
     def __init__(self, name: str, description: str, parameters: BaseModel = EmptyParameters, dangerous=False):
          if name is None:
               raise Exception('name is required')
          if description is None:
               raise Exception('description is required')
          if parameters is None:
               parameters = EmptyParameters

          self.name = name
          self.description = description
          self.parameters = parameters
          self.dangerous = dangerous
    
     def get_tool_function(self):
          return {
               "name": self.name,
               "description": self.description,
               "parameters": self.parameters.schema()
          }

     def run(self, arguments):
          if self.parameters == InputParameters and 'input' in arguments:
               return arguments['input']
               
    


from bondai.prompt.steps_formatter import format_previous_steps

QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about a file. Use the provided information to answer the user's QUESTION at the very end."

def build_prompt(task, steps):
    return f"""

# Task #

This is the task the AI agent is working on:
{task}


# Previous Work #

{steps}


IMPORTANT: You are being asked to analyze the work of an AI agent. 
You must rewrite the  Previous Work section, removing any unnecessary information. 
Try to reduce the amount of text but don't remove too much or the AI agent won't be able to continue working. 
It is VERY important that you make sure the agent still has enough information about each step to continue working.
"""

def compress_steps(task, steps, model):
    formatted_steps = format_previous_steps(steps)
    prompt = build_prompt(task, formatted_steps)
    response = model.get_completion(prompt, QUERY_SYSTEM_PROMPT)[0]
    split = response.split('# Previous Work #')
    
    result = split[1] if len(split) > 1 else split[0]
    return result



from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'file_write'
TOOL_DESCRIPTION = "This tool will overwrite the contents of a file with the text you specify. Just specify the filename of the file using the 'filename' parameter and the text you would like to write to the file with the 'text' parameter."

class Parameters(BaseModel):
    filename: str
    text: str
    thought: str

class FileWriteTool(Tool):
    def __init__(self):
        super(FileWriteTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        filename = arguments.get('filename')
        text = arguments.get('text')

        if filename is None:
            raise Exception('filename is required')
        if text is None:
            raise Exception('text is required')

        with open(filename, 'w') as f:
            f.write(text)
            return f"File {filename} written successfully"




import pypdf
from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'file_read'
TOOL_DESCRIPTION = "This tool will return the contents of a file for you to view. Just specify the filename of the file using the 'filename' parameter."

def is_pdf(filename):
    with open(filename, 'rb') as file:
        header = file.read(4)
    return header == b'%PDF'

def extract_text_from_pdf(file_path):
    with open(file_path, 'rb') as file:
        pdf = pypdf.PdfReader(file)
        text = ''
        for page_number in range(len(pdf.pages)):
            page = pdf.pages[page_number]
            text += page.extract_text()
        return text

class Parameters(BaseModel):
    filename: str
    thought: str

class FileReadTool(Tool):
    def __init__(self):
        super(FileReadTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        filename = arguments.get('filename')
        if filename is None:
            raise Exception('filename is required')

        if is_pdf(filename):
            return extract_text_from_pdf(filename)
        else:
            with open(filename, 'r') as f:
                return f.read()




import os
from termcolor import cprint
from .langchain_tools import build_langchain_tools, is_langchain_installed
from bondai.tools import HumanTool
from bondai.tools.agent import AgentTool
from bondai.tools.alpaca import CreateOrderTool, GetAccountTool, ListPositionsTool
from bondai.tools.file import FileQueryTool, FileReadTool, FileWriteTool
from bondai.tools.gmail import ListEmailsTool, QueryEmailsTool
from bondai.tools.search import GoogleSearchTool, DuckDuckGoSearchTool
from bondai.tools.website import (
    DownloadFileTool,
    WebsiteQueryTool,
)

def build_tool_options():
    tool_options = [
        DownloadFileTool(),
        FileQueryTool(),
        FileWriteTool(), 
        WebsiteQueryTool(),
    ]

    if os.environ.get('ALPACA_API_KEY') and os.environ.get('ALPACA_SECRET_KEY'):
        tool_options.append(CreateOrderTool())
        tool_options.append(GetAccountTool())
        tool_options.append(ListPositionsTool())
    else:
        cprint("Skipping Alpaca Markets tools because ALPACA_API_KEY and ALPACA_SECRET_KEY environment variables are not set.", "yellow")

    if os.environ.get('GOOGLE_API_KEY') and os.environ.get('GOOGLE_CSE_ID'):
        tool_options.append(GoogleSearchTool())
    else:
        tool_options.append(DuckDuckGoSearchTool())
        cprint("Skipping Google Search tool because GOOGLE_API_KEY and GOOGLE_CSE_ID environment variables are not set.", "yellow")

    if 'gmail-token.pickle' in os.listdir():
        tool_options.append(ListEmailsTool())
        tool_options.append(QueryEmailsTool())
    else:
        cprint("Skipping Gmail tools because gmail-token.pickle file is not present.", "yellow")
    
    if is_langchain_installed():
        cprint("Loading LangChain tools...", "yellow")
        langchain_tools = build_langchain_tools()
        for t in langchain_tools:
            tool_options.append(t)
        cprint("Done loading LangChain tools", "yellow")

    return tool_options




import requests
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.util import get_website_links

TOOL_NAME = 'website_extract_hyperlinks'
TOOL_DESCRIPTION = "This tool allows will extract a list of hyperlinks from a website. Just specify the url of the website using the 'url' parameter."
REQUEST_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
}

class Parameters(BaseModel):
    url: str
    thought: str

class WebsiteExtractHyperlinksTool(Tool):
    def __init__(self):
        super(WebsiteExtractHyperlinksTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        url = arguments['url']
        if url is None:
            raise Exception('url is required')

        try:
            links = get_website_links(url)
            return '\n'.join([f"[{a.text}]({a.get('href', '')})" for a in links])
        except requests.Timeout:
            return "The request timed out."



import requests
from bs4 import BeautifulSoup

REQUEST_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
}

def get_website_html(url):
        response = requests.get(url, headers=REQUEST_HEADERS, timeout=10)
        return response.text

def get_html_text(html):
        soup = BeautifulSoup(html, "html.parser")
        return soup.get_text()

def get_website_text(url):
        html = get_website_html(url)
        return get_html_text(html)

def query_website_html(url, xpath):
        html = get_website_html(url)
        soup = BeautifulSoup(html, "html.parser")
        root = soup.html
        return root.xpath(xpath)

def get_website_links(url):
        html = get_website_html(url)
        soup = BeautifulSoup(html, "html.parser")
        return soup.find_all("a")



from pydantic import BaseModel
from bondai.tools.tool import Tool
from duckpy import Client

MAX_RESULT_COUNT = 20
DEFAULT_RESULT_COUNT = 5
TOOL_NAME = 'duck_duck_go_search'
TOOL_DESCRIPTION = "This tool allows you to search DuckDuckGo. Specify your search string in the 'query' parameter and it will return a list that includes the title and url of matched websites."
TOOL_DESCRIPTION = f"This tool allows you to retrieve a paginated list of search results. You must specify your search string in the 'query' parameter. You can specify the number of search results to return by setting the 'count' parameter. The maximum count is {MAX_RESULT_COUNT} and the default is {DEFAULT_RESULT_COUNT}. To paginate through the full list of all search results just increment the 'page' parameter. By default 'page' is set to 1."

class Parameters(BaseModel):
    query: str
    count: int
    page: int
    thought: str

def search_duckduckgo(query, count=10, page=1):
    ddg_client = Client()
    response = ddg_client.search(query, count=count, page=page)
    results =  '\n'.join([f"[{r.title}]: ({r.url})" for r in response])

    if len(results) > count:
        results = results[:count]
    
    return results

class DuckDuckGoSearchTool(Tool):
    def __init__(self):
        super(DuckDuckGoSearchTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        query = arguments.get('query')
        count = int(arguments.get('count', '5'))
        page = int(arguments.get('page', '1'))

        if query is None:
            raise Exception('query is required')

        if count > MAX_RESULT_COUNT:
            count = MAX_RESULT_COUNT

        output = search_duckduckgo(query, count, page)
        print(output)
        return output

    




import json

def format_step(step, count):
    str_work = ''
    if step.function:
        function_input = ...
        try:
            args = json.loads(step.function['arguments'])
            function_input = ''
            for k in args.keys():
                function_input += f"{k}: {args[k]}\n"
        except Exception:
            pass

        if function_input:
            str_work += f"""

## Step {count} ##
In Step #{count} you used the {step.function['name']} function. You used these parameters:
{function_input}

The following results were returned:
{step.output}
"""
        else:
            str_work += f"""
## Step {count} ##
In Step #{count} you used the {step.function['name']} function. The following results were returned:
{step.output}
"""
    else:
        str_work += f"""

## Step {count} ##
In Step #{count} you made a mistake and did not specify a function to use. It is important that you select a function.
"""

    return str_work

def format_previous_steps(model, previous_steps, max_tokens=None):
    str_work = ''
    for i, s in enumerate(previous_steps):
        str_step = format_step(s, i+1)
        if not max_tokens or model.count_tokens(str_work + str_step) <= max_tokens:
            str_work += str_step
        else:
            break
    return str_work


from bondai.models import EmbeddingModel
from .openai_models import MODEL_NAMES, MODELS, MODEL_TYPE_EMBEDDING
from .openai_wrapper import create_embedding, count_tokens, get_max_tokens

class OpenAIEmbeddingModel(EmbeddingModel):
    def __init__(self, model):
        if model not in MODEL_NAMES:
            raise Exception(f"Model {model} is not supported.")
        if MODELS[model]['model_type'] != MODEL_TYPE_EMBEDDING:
            raise Exception(f"Model {model} is not an embedding model.")
        self.model = model

    def create_embedding(self, prompt):
        return create_embedding(prompt, self.model)

    def count_tokens(self, prompt):
        return count_tokens(prompt, self.model)

    def get_max_tokens(self):
        return get_max_tokens(self.model)


from bondai import Agent
from bondai.tools import Tool, InputParameters
from bondai.models.openai import OpenAILLM, MODEL_GPT4_0613

TOOL_NAME = 'agent_tool'
TOOL_DESCRIPTION = "This tool allows you to delegate tasks to other agents. This can be really helpful for taking a complex task and breaking it down into smaller, more manageable pieces. Just include a highly descriptive prompt in the 'input' parameter for this task. The more detailed your description the better the agent will be at the task. If you would like to create multiple agents just pass an array of prompts in the 'input' parameter."

class AgentTool(Tool):
    def __init__(self, prompt_builder=None, tools=[], llm=OpenAILLM(MODEL_GPT4_0613)):
        super(AgentTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, InputParameters)
        self.agent = Agent(prompt_builder, tools=tools, llm=llm)
    
    def run(self, arguments):
        agent_prompt = arguments['input']
        if isinstance(agent_prompt, list):
            output = ''
            for i,p in enumerate(agent_prompt):
                result = self.agent.run(p)
                output += f"Agent #{i+1} Response: {result.output}\n"
            
            return output
        else:
            return self.agent.run(agent_prompt).output



import os
import pkg_resources
from datetime import datetime
from bondai.prompt import PromptBuilder
from bondai.prompt.steps_formatter import format_previous_steps

DEFAULT_PROMPT_TEMPLATE_FILENAME='default_prompt_template.md'

current_dir = os.path.dirname(os.path.abspath(__file__))
prompt_template_path = os.path.join(current_dir, DEFAULT_PROMPT_TEMPLATE_FILENAME)

if os.path.exists(prompt_template_path):
    with open(prompt_template_path, 'r') as file:
        DEFAULT_PROMPT_TEMPLATE = file.read()
else:
    DEFAULT_PROMPT_TEMPLATE = pkg_resources.resource_string(__name__, f"prompt/{DEFAULT_PROMPT_TEMPLATE_FILENAME}").decode()

class DefaultPromptBuilder(PromptBuilder):

    def __init__(self, llm, prompt_template=DEFAULT_PROMPT_TEMPLATE):
        self.llm = llm
        self.prompt_template = prompt_template

    def build_prompt(self, task, tools, previous_steps=[], max_tokens=None):
        if not max_tokens:
            max_tokens = self.llm.get_max_tokens()

        prompt = self.prompt_template.replace('{DATETIME}', str(datetime.now()))
        prompt = prompt.replace('{TASK}', task)

        if len(previous_steps) > 0:
            str_work = 'This is a list of previous steps that you already completed on this TASK.'
            
            remaining_tokens = max_tokens - self.llm.count_tokens(prompt)
            str_work += format_previous_steps(self.llm, previous_steps, remaining_tokens)
        else:
            str_work = '**No previous steps have been completed**'

        prompt = prompt.replace('{WORK}', str_work)
        return prompt



import uuid
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.util import semantic_search
from bondai.models.openai import (
    OpenAILLM, 
    OpenAIEmbeddingModel, 
    MODEL_GPT35_TURBO_16K, 
    MODEL_TEXT_EMBEDDING_ADA_002
)

TOOL_NAME = 'response_query'
QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about a file. Use the provided information to answer the user's QUESTION at the very end."
TOOL_DESCRIPTION = "This tool allows you to ask a questions about responses that are too large and contain too many tokens. Just specify the response_id using the 'response_id' parameter and specify your question using the 'question' parameter."

def build_prompt(question, context):
    return f"""{context}


IMPORTANT: Answer the following question for the user.
QUESTION: {question}
"""

class Parameters(BaseModel):
    response_id: str
    question: str
    thought: str

class ResponseQueryTool(Tool):
    def __init__(self, llm=OpenAILLM(MODEL_GPT35_TURBO_16K), embedding_model=OpenAIEmbeddingModel(MODEL_TEXT_EMBEDDING_ADA_002)):
        super(ResponseQueryTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.llm = llm
        self.embedding_model = embedding_model
        self.responses = {}
    
    def add_response(self, response):
        response_id = str(uuid.uuid4())
        self.responses[response_id] = response
        return response_id
    
    def run(self, arguments):
        response_id = arguments['response_id']
        question = arguments['question']

        if response_id is None:
            raise Exception('response_id is required')
        if question is None:
            raise Exception('question is required')

        if response_id in self.responses:
            text = self.responses[response_id]
            text = semantic_search(self.embedding_model, question, text, 16000)
            prompt = build_prompt(question, text)
            response = self.llm.get_completion(prompt, QUERY_SYSTEM_PROMPT, model=self.model)[0]
            return response
        else:
            return f"{response_id} is not a valid response_id"



from bondai.tools import Tool, InputParameters

class LangChainTool(Tool):
    def __init__(self, tool, parameters=InputParameters, dangerous=False):
        super(LangChainTool, self).__init__(tool.name, tool.description, parameters, dangerous=dangerous)
        self.tool = tool

    def run(self, arguments):
        return self.tool.run(arguments)

    




import requests
from pydantic import BaseModel
from bondai.tools import Tool

TOOL_NAME = 'download_file'
TOOL_DESCRIPTION = "This tool allows to you to download a file. Just provide the url to the file in the 'url' parameter and the filename it should be saved to in the 'filename' parameter."
REQUEST_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
}

class Parameters(BaseModel):
    url: str
    filename: str
    thought: str

class DownloadFileTool(Tool):
    def __init__(self):
        super(DownloadFileTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        url = arguments.get('url')
        filename = arguments.get('filename')

        if url is None:
            raise Exception('url is required')
        if filename is None:
            raise Exception('filename is required')

        try:
            response = requests.get(url)

            with open(filename, 'wb') as output_file:
                output_file.write(response.content)
            
            return f"The file was successfully downloaded to {filename}."
        except requests.Timeout:
            return "The request timed out."




import json
import time
import openai
import tiktoken
from .openai_models import MODELS, MODEL_TYPE_LLM

embedding_tokens = 0
embedding_costs = 0.0
gpt_tokens = 0
gpt_costs = 0.0

logger = None

def enable_logging(model_logger):
    global logger
    logger = model_logger

def disable_logging():
    global logger
    logger = None

def get_gpt_tokens():
    return gpt_tokens

def get_embedding_tokens():
    return embedding_tokens

def get_gpt_costs():
    return gpt_costs

def get_embedding_costs():
    return embedding_costs

def get_total_cost():
    return embedding_costs + gpt_costs

def reset_total_cost():
    global embedding_costs, embedding_tokens, gpt_costs, gpt_tokens
    embedding_costs = 0.0
    embedding_tokens = 0
    gpt_costs = 0.0
    gpt_tokens = 0

def calculate_cost(model_name, usage):
    global embedding_costs, embedding_tokens, gpt_costs, gpt_tokens

    if model_name in MODELS:
        model = MODELS[model_name]
        token_count = usage['total_tokens']

        if model['model_type'] == MODEL_TYPE_LLM:
            gpt_tokens += token_count
            gpt_costs += (usage['prompt_tokens'] * model['input_price_per_token']) + (usage['completion_tokens'] * model['output_price_per_token'])
        else:
            embedding_tokens += token_count
            embedding_costs += token_count * model['price_per_token']
    else:
        print(f"Unknown model: {model_name}")


def get_max_tokens(model):
    return MODELS[model]['max_tokens']


def count_tokens(prompt, model):
    encoding = tiktoken.encoding_for_model(model)
    return len(encoding.encode(prompt))

def create_embedding(text, model="text-embedding-ada-002"):
    tries = 0
    while True:
        try:
            response = openai.Embedding.create(input=[text], model=model)
            calculate_cost(model, response['usage'])
            return response["data"][0]["embedding"]
        except Exception as e:
            if tries >= 3:
                raise e
            
            print(e)
            time.sleep(5)
            tries += 1

def get_completion(prompt, system_prompt='', previous_messages=[], functions=[], model='gpt-4'):
    global logger
    prompt_log = ''
    messages = []

    if len(functions) > 0:
        fs_str = json.dumps(functions)
        prompt_log += f"TOOLS:\n{fs_str}\n\n"

    if system_prompt:
        prompt_log += f"SYSTEM: {system_prompt}\n\n"
        messages.append({
            "role": "system",
            "content": system_prompt
        })
    
    if len(previous_messages) > 0:
        prompt_log += "PREVIOUS MESSAGES:\n"
        for m in previous_messages:
            prompt_log += f"{m}\n"
            messages.append({
                "role": "user",
                "content": m
            })
        prompt_log += '\n'

    prompt_log += f"PROMPT: {prompt}\n\n"
    messages.append({
        "role": "user",
        "content": prompt
    })

    attempts = 0
    max_retries = 3
    success = False
    while not success:
        try:
            attempts += 1
            if len(functions) > 0:
                response = openai.ChatCompletion.create(
                    model=model,
                    messages=messages,
                    functions=functions,
                    temperature=0.1,
                )
            else:
                response = openai.ChatCompletion.create(
                    model=model,
                    messages=messages,
                    temperature=0.1,
                )
            
            success = True
        except Exception as e:
            time.sleep(15)
            if attempts >= max_retries:
                raise e
    
    calculate_cost(model, response['usage'])
    message = response["choices"][0]["message"]

    if message.get("function_call"):
        function = message["function_call"]
        if logger:
            try:
                logger.log(prompt_log, message["content"], {
                    'name': function['name'],
                    'arguments': json.loads(function['arguments'])
                })
            except json.decoder.JSONDecodeError:
                pass
        return message["content"], function
    else:
        if logger:
            logger.log(prompt_log, message["content"])
        return message["content"], None



import requests
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.util import get_website_text, semantic_search
from bondai.models.openai import (
    OpenAILLM, 
    OpenAIEmbeddingModel, 
    MODEL_GPT35_TURBO_16K, 
    MODEL_TEXT_EMBEDDING_ADA_002
)

TOOL_NAME = 'website_query'
QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about a website. Use the provided information to answer the user's QUESTION at the very end."
TOOL_DESCRIPTION = "This tool allows to ask a question about the text content of any website including summarization. Just specify the url of the website using the 'url' parameter and specify your question using the 'question' parameter."
REQUEST_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
}

def build_prompt(question, context):
    return f"""{context}


IMPORTANT: Using the information provided above, answer the following question for the user.
QUESTION: {question}
"""

class Parameters(BaseModel):
    url: str
    question: str
    thought: str

class WebsiteQueryTool(Tool):
    def __init__(self, llm=OpenAILLM(MODEL_GPT35_TURBO_16K), embedding_model=OpenAIEmbeddingModel(MODEL_TEXT_EMBEDDING_ADA_002)):
        super(WebsiteQueryTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.llm = llm
        self.embedding_model = embedding_model
    
    def run(self, arguments):
        url = arguments['url']
        question = arguments['question']

        if url is None:
            raise Exception('url is required')
        if question is None:
            raise Exception('question is required')

        try:
            text = get_website_text(url)
        except requests.Timeout:
            return "The request timed out."

        text = semantic_search(self.embedding_model, question, text, 16000)
        prompt = build_prompt(question, text)
        response = self.llm.get_completion(prompt, QUERY_SYSTEM_PROMPT)[0]

        return response




import os
from googleapiclient.discovery import build
from pydantic import BaseModel
from bondai.tools.tool import Tool

MAX_RESULT_COUNT = 20
DEFAULT_RESULT_COUNT = 5
TOOL_NAME = 'google_search'
TOOL_DESCRIPTION = "This tool allows you to search Google. Specify your search string in the 'query' parameter and it will return a list that includes the title and url of matched websites."
TOOL_DESCRIPTION = f"This tool allows you to retrieve a paginated list of google search results. You must specify your search string in the 'query' parameter. You can specify the number of search results to return by setting the 'count' parameter. The maximum count is {MAX_RESULT_COUNT} and the default is {DEFAULT_RESULT_COUNT}. To paginate through the full list of all search results just increment the 'page' parameter. By default 'page' is set to 1."

GOOGLE_API_KEY = os.environ.get('GOOGLE_API_KEY')
GOOGLE_CSE_ID = os.environ.get('GOOGLE_CSE_ID')

class Parameters(BaseModel):
    query: str
    count: int
    page: int
    thought: str

class GoogleSearchTool(Tool):
    def __init__(self, google_api_key=GOOGLE_API_KEY, google_cse_id=GOOGLE_CSE_ID):
        super(GoogleSearchTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.google_cse_id = google_cse_id
        self.siterestrict = False
        self.search_engine = build("customsearch", "v1", developerKey=google_api_key)
    
    def run(self, arguments):
        query = arguments.get('query')
        count = int(arguments.get('count', '5'))
        page = int(arguments.get('page', '1'))

        if query is None:
            raise Exception('query is required')

        if count > MAX_RESULT_COUNT:
            count = MAX_RESULT_COUNT

        cse = self.search_engine.cse()
        if self.siterestrict:
            cse = cse.siterestrict()
        
        start = (page-1) * count
        res = cse.list(q=query, cx=self.google_cse_id, start=start, num=count).execute()
        items = res.get("items", [])
        
        output = ''
        for i in items:
            output += f"[{i['title']}]({i['link']})\n"

        return output

    




from .model_logger import ModelLogger
from .semantic_search import semantic_search
from .web import get_website_html, get_html_text, get_website_text, query_website_html, get_website_links

__all__ = [
    "ModelLogger",
    "semantic_search",
    "get_website_html",
    "get_html_text",
    "get_website_text",
    "query_website_html",
    "get_website_links"
]


from .create_order import CreateOrderTool
from .get_account import GetAccountTool
from .list_positions import ListPositionsTool
from .response_formatter import format_order_response, format_account_response, format_positions_response

__all__ = [
    "CreateOrderTool",
    "GetAccountTool",
    "ListPositionsTool",
    "format_order_response",
    "format_account_response",
    "format_positions_response"
]



import os
from pydantic import BaseModel
from bondai.tools import Tool
from .response_formatter import format_order_response
from alpaca.trading.client import TradingClient
from alpaca.trading.requests import MarketOrderRequest, LimitOrderRequest
from alpaca.common.exceptions import APIError

TOOL_NAME = 'create_order'
TOOL_DESCRIPTION = """This tool allows you to create an Order to buy or sell a stock. 
When using this tool, you must specify either 'buy' or 'sell' for the 'side' parameter. 
You must also specify the 'symbol' parameter which is the stock symbol of the stock you want to buy or sell. 
You must also specify the 'quantity' parameter which is the number of shares you want to buy or sell. 
The 'order_type' parameter is optional. It will default to a 'market' order by you can also specify a 'limit' order.
The 'limit_price' parameter is only required if you specify a 'limit' order.
The 'time_in_force' parameter is optional. It will default to 'day' but you can also specify 'gtc', 'opg', 'cls', 'ioc', or 'fok'."""

ALPACA_API_KEY = os.environ.get('ALPACA_API_KEY')
ALPACA_SECRET_KEY = os.environ.get('ALPACA_SECRET_KEY')

class Parameters(BaseModel):
    side: str
    symbol: str
    quantity: str
    order_type: str
    time_in_force: str
    limit_price: str
    thought: str

class CreateOrderTool(Tool):
    def __init__(self, alpaca_api_key=ALPACA_API_KEY, alpaca_secret_key=ALPACA_SECRET_KEY):
        super(CreateOrderTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.trading_client = TradingClient(alpaca_api_key, alpaca_secret_key, paper=True)
    
    def run(self, arguments):
        side = arguments.get('side')
        symbol = arguments.get('symbol')
        quantity = arguments.get('quantity')
        order_type = arguments.get('order_type', 'market')
        time_in_force = arguments.get('time_in_force', 'day')
        limit_price = arguments.get('limit_price', None)

        if not side in ['buy', 'sell']:
            return 'Invalid side. Must be either "buy" or "sell".'
        if not symbol:
            return 'Invalid symbol.'
        if not quantity:
            return 'Invalid quantity.'
        if not order_type in ['market', 'limit']:
            return 'Invalid order type. Must be either "market" or "limit".'
        if not time_in_force in ['day', 'gtc', 'opg', 'cls', 'ioc', 'fok']:
            return 'Invalid time in force. Must be either "day", "gtc", "opg", "cls", "ioc", or "fok".'
        if order_type == 'limit' and not limit_price:
            return 'Invalid limit price.'

        if order_type == 'market':
            order = MarketOrderRequest(symbol=symbol, qty=quantity, side=side, time_in_force=time_in_force)
        else:
            order = LimitOrderRequest(symbol=symbol, limit_price=limit_price, qty=quantity, side=side, time_in_force=time_in_force)

        try:
            response = self.trading_client.submit_order(order_data=order)
            return format_order_response(response)
        except APIError as e:
            return e.message




import nltk
import numpy as np

nltk.download("punkt", quiet=True)

MAX_EMBED_TOKENS = 250
SENTENCE_CONCAT_COUNT = 4

def cosine_similarity(vec1, vec2):
  dot_product = np.dot(vec1, vec2)
  norm1 = np.linalg.norm(vec1)
  norm2 = np.linalg.norm(vec2)
  return dot_product / (norm1 * norm2)

def split_text(embedding_model, text, max_length):
    result = []
    split = nltk.sent_tokenize(text)
    split = concatenate_strings(split, SENTENCE_CONCAT_COUNT)
    for s in split:
        if embedding_model.count_tokens(s) > max_length:
            split2 = s.split('\n')
            for s2 in split2:
                if embedding_model.count_tokens(s2) > max_length:
                    split3 = split_tokens(embedding_model, s2, max_length)
                    for s3 in split3:
                        result.append(s3)
                else:
                    result.append(s2)
        else:
            result.append(s)

    filtered_list = [s for s in result if s.strip()]
    return filtered_list

def concatenate_strings(arr, n):
    arr2 = []
    for i in range(0, len(arr), n):
        concat_str = ''.join(arr[i: i+n])
        arr2.append(concat_str)
    return arr2

def split_tokens(embedding_model, input, max_length):
    if embedding_model.count_tokens(input) <= max_length: return input

    item = ''
    result = []
    for i, c in enumerate(input):
        item += c
        if embedding_model.count_tokens(item) >= max_length:
            result.append(item)
            item = ''
    
    return result


def semantic_search(embedding_model, query, text, max_tokens):
    if embedding_model.count_tokens(text) <= max_tokens: return text

    results = []
    sentences = split_text(embedding_model, text, MAX_EMBED_TOKENS)
    query_embedding = embedding_model.create_embedding(query)

    for i,s in enumerate(sentences):
        try:
            s_embedding = embedding_model.create_embedding(s)
            similarity = cosine_similarity(query_embedding, s_embedding)

            results.append({
                "sentence": s,
                "similarity": similarity,
                "order": i
            })
        except Exception as e:
            print(e)
    
    results.sort(key=lambda x: x['similarity'], reverse=True)

    filtered = []
    str_items = ''
    for r in results:
        if embedding_model.count_tokens(f"{str_items}\n\n{r['sentence']}") <= max_tokens:
            str_items += f"\n\n{r['sentence']}"
            filtered.append(r)
        else:
            break

    filtered.sort(key=lambda x: x['order'])
    filtered_str = map(lambda x: x['sentence'], filtered)
    output = '\n\n'.join(filtered_str)

    return output

    


import requests
from pydantic import BaseModel
from bondai.tools.tool import Tool
from bondai.util.web import get_website_html
from bondai.models.openai import (
    OpenAILLM,  
    MODEL_GPT35_TURBO_16K
)

TOOL_NAME = 'website_html_query'
QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about the HTML in a website. Use the provided HTML content to answer the user's QUESTION at the very end."
TOOL_DESCRIPTION = "This tool allows to ask a question about the raw HTML content of a website. Just specify the url of the website using the 'url' parameter and specify your question using the 'question' parameter."

def build_prompt(question, context):
    return f"""{context}


IMPORTANT: Using the information provided above, answer the following question for the user.
QUESTION: {question}
"""

class Parameters(BaseModel):
    url: str
    question: str
    thought: str

class WebsiteHtmlQueryTool(Tool):
    def __init__(self, llm=OpenAILLM(MODEL_GPT35_TURBO_16K)):
        super(WebsiteHtmlQueryTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.llm = llm
    
    def run(self, arguments):
        url = arguments['url']
        question = arguments['question']

        if url is None:
            raise Exception('url is required')
        if question is None:
            raise Exception('question is required')

        try:
            html = get_website_html(url)
        except requests.Timeout:
            return "The request timed out."

        prompt = build_prompt(question, html)
        response = self.llm.get_completion(prompt, QUERY_SYSTEM_PROMPT, model=self.model)[0]

        return response




import json
from pydantic import BaseModel
from typing import List
from bondai.tools import Tool

TOOL_NAME = 'final_answer'
TOOL_DESCRIPTION = """Use this tool ONLY after you have captured ALL of the required information from the user. This tool will send the information to the next AI assistant that will complete the task.
- user_exit: This is a boolean value that indicates whether the user has asked to exit. If the user has asked to exit you should NOT ask for any other information from the user.
- task_description: This must be detailed enough for the next AI assistant to understand what the user wants to do. Ask the user any necessary follow up questions.
- task_budget: This is the maximum amount of money the user is willing to spend on this task. Ask the user any necessary follow up questions.
- tool_ids: This is a list of the IDs for the tools that the next AI assistant will need to complete the task. Ask the user any necessary follow up questions.
- user_confirmation: This is a boolean value that indicates whether the user has confirmed the task description, task budget, and tool ids.
"""

class Parameters(BaseModel):
    user_exit: bool = False
    task_description: str
    task_budget: float
    user_confirmation: bool = False

class OnboardingTool(Tool):
    def __init__(self):
        super(OnboardingTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
    
    def run(self, arguments):
        user_exit = arguments.get('user_exit')
        if user_exit:
            return json.dumps({
                'user_exit': True
            })

        task_description = arguments.get('task_description')
        task_budget = arguments.get('task_budget')
        user_confirmation = arguments.get('user_confirmation')

        if not task_description:
            raise Exception('You must provide a task description.')
        if not task_budget:
            raise Exception('You must provide a task budget.')
        if not user_confirmation:
            raise Exception('You must confirm the task description, task budget, and tool ids with the user before calling the final_answer tool.')
        
        return json.dumps({
            'task_description': task_description,
            'task_budget': task_budget
        })



import os
from bondai.tools.tool import Tool
from .response_formatter import format_positions_response
from alpaca.trading.client import TradingClient

TOOL_NAME = 'list_positions'
TOOL_DESCRIPTION = "This tool will list all of your currently open positions."

ALPACA_API_KEY = os.environ.get('ALPACA_API_KEY')
ALPACA_SECRET_KEY = os.environ.get('ALPACA_SECRET_KEY')

class ListPositionsTool(Tool):
    def __init__(self, alpaca_api_key=ALPACA_API_KEY, alpaca_secret_key=ALPACA_SECRET_KEY):
        super(ListPositionsTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION)
        self.trading_client = TradingClient(alpaca_api_key, alpaca_secret_key, paper=True)
    
    def run(self, arguments):
        response = self.trading_client.get_all_positions()
        return format_positions_response(response)




from setuptools import setup, find_packages

# If you have a requirements.txt, you can read it to set the install_requires parameter.
with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

setup(
    name="bondai",
    version="0.1.1",
    description="An AI-powered console assistant with a versatile API for seamless integration into applications.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author="Kevin Rohling",
    author_email="kevin@kevinrohling.com",
    url="https://github.com/krohling/bondai",
    packages=find_packages(),  # This will include all packages under the bondai directory
    scripts=['scripts/bondai'],
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
    ],
)



from .openai_llm import OpenAILLM
from .openai_embedding_model import OpenAIEmbeddingModel
from .openai_wrapper import get_total_cost, reset_total_cost, enable_logging, disable_logging
from .openai_models import (
    MODEL_GPT4, 
    MODEL_GPT4_0613, 
    MODEL_GPT4_32K, 
    MODEL_GPT35_TURBO,
    MODEL_GPT35_TURBO_16K,
    MODEL_GPT35_TURBO_0613,
    MODEL_GPT35_TURBO_16K_0613,
    MODEL_TEXT_EMBEDDING_ADA_002
)

__all__ = [
    'OpenAILLM',
    'OpenAIEmbeddingModel',
    'get_total_cost',
    'reset_total_cost',
    'enable_logging',
    'disable_logging',
    'MODEL_GPT4',
    'MODEL_GPT4_0613',
    'MODEL_GPT4_32K',
    'MODEL_GPT35_TURBO',
    'MODEL_GPT35_TURBO_16K',
    'MODEL_GPT35_TURBO_0613',
    'MODEL_GPT35_TURBO_16K_0613',
    'MODEL_TEXT_EMBEDDING_ADA_002',
]


from .tool import Tool, EmptyParameters, InputParameters
from .human_tool import HumanTool
from .langchain_tool import LangChainTool
from .response_query import ResponseQueryTool

__all__ = [
    "Tool",
    "EmptyParameters",
    "InputParameters",
    "HumanTool",
    "LangChainTool"
    "ResponseQueryTool"
]


from termcolor import cprint
from bondai.tools import Tool, InputParameters

TOOL_NAME = 'human'
TOOL_DESCRIPTION = "If you find your are stuck and not making progress you can use this tool to ask a human a question. The human cannot perform actions for you but they can give you ideas for your next step. Only ask a human if you are stuck and not sure how to proceed. Just add your question to the 'input' parameter."

class HumanTool(Tool):
    def __init__(self):
        super(HumanTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, InputParameters)
    
    def run(self, arguments):
        question = arguments['input']

        if question is None:
            raise Exception('input is required')

        cprint(f"\n{question}\n", 'white', attrs=["bold"])
        output = input()

        return output



from abc import ABC, abstractmethod

class EmbeddingModel(ABC):

    @abstractmethod
    def create_embedding(prompt):
        pass

    @abstractmethod
    def count_tokens(prompt):
        pass

    @abstractmethod
    def get_max_tokens():
        pass


from .extract_hyperlinks import WebsiteExtractHyperlinksTool
from .html_query import WebsiteHtmlQueryTool
from .query_agent import WebsiteQueryAgentTool
from .query import WebsiteQueryTool
from .download_file import DownloadFileTool


__all__ = [
    "WebsiteExtractHyperlinksTool",
    "WebsiteHtmlQueryTool",
    "WebsiteQueryAgentTool",
    "WebsiteQueryTool",
    "DownloadFileTool"
]



import pypdf
from pydantic import BaseModel
from bondai.tools import Tool
from bondai.util import semantic_search
from bondai.models.openai import (
    OpenAILLM, 
    OpenAIEmbeddingModel, 
    MODEL_GPT35_TURBO_16K, 
    MODEL_TEXT_EMBEDDING_ADA_002
)

TOOL_NAME = 'file_query'
QUERY_SYSTEM_PROMPT = "You are a helpful question and answer assistant designed to answer questions about a file. Use the provided information to answer the user's QUESTION at the very end."
TOOL_DESCRIPTION = "This tool allows you to ask a question about the text content of any file including summarization. Just specify the filename of the file using the 'filename' parameter and specify your question using the 'question' parameter."

def is_pdf(filename):
    with open(filename, 'rb') as file:
        header = file.read(4)
    return header == b'%PDF'

def extract_text_from_pdf(file_path):
    with open(file_path, 'rb') as file:
        pdf = pypdf.PdfReader(file)
        text = ''
        for page_number in range(len(pdf.pages)):
            page = pdf.pages[page_number]
            text += page.extract_text()
        return text


def build_prompt(question, context):
    return f"""{context}


IMPORTANT: Answer the following question for the user.
QUESTION: {question}
"""

class Parameters(BaseModel):
    filename: str
    question: str
    thought: str

class FileQueryTool(Tool):
    def __init__(self, llm=OpenAILLM(MODEL_GPT35_TURBO_16K), embedding_model=OpenAIEmbeddingModel(MODEL_TEXT_EMBEDDING_ADA_002)):
        super(FileQueryTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.llm = llm
        self.embedding_model = embedding_model
    
    def run(self, arguments):
        filename = arguments.get('filename')
        question = arguments.get('question')

        if filename is None:
            raise Exception('filename is required')
        if question is None:
            raise Exception('question is required')

        if is_pdf(filename):
            text = extract_text_from_pdf(filename)
        else:
            with open(filename, 'r') as f:
                text = f.read()

        text = semantic_search(self.embedding_model, question, text, 16000)
        prompt = build_prompt(question, text)
        response = self.llm.get_completion(prompt, QUERY_SYSTEM_PROMPT)[0]

        return response




import os
import json
from datetime import datetime

def get_instance_dir(logging_dir):
    dir_name = datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')

    path = f"{logging_dir}/{dir_name}"
    if not os.path.exists(path):
        os.makedirs(path)
    
    return path

def write_file(filename, content):
    with open(filename, 'w') as f:
        f.write(content)

class ModelLogger:
    def __init__(self, logging_dir='./logs'):
        self.logging_dir = logging_dir
    
    def log(self, prompt, response, function=None):
        instance_path = get_instance_dir(self.logging_dir)

        write_file(f"{instance_path}/prompt.txt", prompt)
        if response:
            write_file(f"{instance_path}/response.txt", response)
        if function:
            f_str = json.dumps(function)
            write_file(f"{instance_path}/function.txt", f_str)
        



from pydantic import BaseModel
from bondai import Agent
from bondai.tools import Tool
from bondai.prompt import DefaultPromptBuilder
from .query import WebsiteQueryTool
from .extract_hyperlinks import WebsiteExtractHyperlinksTool
from .html_query import WebsiteHtmlQueryTool
from .download_file import DownloadFileTool
from bondai.models.openai import (
    OpenAILLM, 
    MODEL_GPT4_0613
)

TOOL_NAME = 'website_query_agent'
TOOL_DESCRIPTION = "This tool allows to ask a question about the text content of any website including summarization. Just specify the url of the website using the 'url' parameter and specify your question using the 'question' parameter."

WEBSITE_TOOLS = [
    WebsiteQueryTool(),
    WebsiteExtractHyperlinksTool(),
    WebsiteHtmlQueryTool(),
    DownloadFileTool()
]

class Parameters(BaseModel):
    url: str
    question: str
    thought: str

def build_prompt(url, question):
    return f"""You are a powerful AI agent that has been tasked with answering a question about a website.
You have access to a number of tools that allow you to access content on websites.
You can use the tools to extract information from the website and then use that information to answer the question.
This is the url of the website:
{url}

This is the question you need to answer: 
{question}"""

class WebsiteQueryAgentTool(Tool):
    def __init__(self, llm=OpenAILLM(MODEL_GPT4_0613), budget=None):
        super(WebsiteQueryAgentTool, self).__init__(TOOL_NAME, TOOL_DESCRIPTION, Parameters)
        self.agent = Agent(prompt_builder=DefaultPromptBuilder(llm), tools=WEBSITE_TOOLS, llm=llm, budget=budget)
    
    def run(self, arguments):
        self.agent.reset_memory()
        url = arguments['url']
        question = arguments['question']

        if url is None:
            raise Exception('url is required')
        if question is None:
            raise Exception('question is required')

        prompt = build_prompt(url, question)

        result = self.agent.run(prompt)
        self.agent.reset_memory()
        return result.output

    

